#!/bin/bash

set -e

JFMC_USER_NAME=${JFMC_USER_NAME:-"jfmc"}
JFMC_USER_GROUP=${JFMC_USER_GROUP:-"jfmc"}
JFMC_USER_ID=${JFMC_USER_ID:-"1050"}
JFMC_USER_GROUP_ID=${JFMC_USER_GROUP_ID:-"1050"}

#Print the input with additional formatting to indicate a section/title
title () {
    echo
    echo "-----------------------------------------------------"
    printf "| %-50s|\n" "$1"
    echo "-----------------------------------------------------"
}

# This function prints the echo with color.
#If invoked with a single string, treats it as INFO level
#Valid inputs for the second parameter are DEBUG and INFO
log() {
    echo ""
    echo -e $1
}

# This function prints the echo with color.
#If invoked with a single string, treats it as INFO level
#Valid inputs for the second parameter are DEBUG and INFO
notice() {
    echo ""
    echo -e "\033[0;36m $1 \033[0m"
    echo ""
}

# Utility method to display warnings and important information
warn() {
    echo ""
    echo -e "\033[33m $1 \033[0m"
    echo ""
}

exitOnError()
{
    msg=$1
    echo ""
    echo -e "\033[31m ERROR: ${msg} \033[0m"
    exit 1
}

# A small timeout ensures that the content printed in the console and the logs remains legible
read_timeout=1

# Labels used in messages displayed to the user
ELASTIC_SEARCH_LABEL='Elasticsearch'
MONGO_LABEL='MongoDB'
POSTGRES_LABEL='Postgres'

# PREREQUISITES
DOCKER_SERVER_MIN_VER='1.11'
#DOCKER_SERVER_CURRENT_VER, this should be passed by the installer script [mission-control.sh] as $(docker version --format '{{.Server.Version}}') || DOCKER_SERVER_CURRENT_VER=""

MAX_OPEN_FILES_MIN_VALUE=10000
MAX_OPEN_FILES_REQUIRED_VALUE=32000

# JFMC_APP_SERVICES_LIST, the services names as they are in the docker-cmpose file, SPACE separated
JFMC_APP_SERVICES_LIST="server insight_server"

# PREREQUISITES
DOCKER_SERVER_MIN_VER='1.11'
#DOCKER_SERVER_CURRENT_VER, this should be passed by the installer script [mission-control.sh] as $(docker version --format '{{.Server.Version}}') || DOCKER_SERVER_CURRENT_VER=""

#Repeated Messages
LOCALHOST_WARNING="Important: When answering the prompts that follow, remember to give a URL that will be reachable from inside other docker containers\n\
For instance, specifying 'localhost' will not work. The default value (docker.for.mac.localhost) will work for MacOS"

aparse()
{
    if [ ! -z $1 ]; then
        command=$1
        shift
    else
        command=usage
        return 0
    fi

    if [[ $# > 0 ]] ; then
              ALL_ARGS="${ALL_ARGS}${1} "
    fi
}

if [ -z "$COMPOSE_PROJECT_NAME" ]; then
    COMPOSE_PROJECT_NAME="jfmc"
fi

DEFAULT_INSTALLER_DATA_FOLDER="/data"
 
# Load default environment and override it with user environment
JFMC_CONF="$DEFAULT_INSTALLER_DATA_FOLDER/$JFMC_CONF_FILE_NAME"
JFMC_CONF_DEFAULT="/opt/jfrog/jfmc/installer/jfmc-env-default.conf"
[ -f "${JFMC_CONF_DEFAULT}" ] && source "${JFMC_CONF_DEFAULT}" || true
[ -f "${JFMC_CONF}" ]         && source "${JFMC_CONF}"         || true

DISPLAY_INSTALL_LOG_LOCATION="$JFMC_MOUNT_ROOT/jfmc-installer"
DISPLAY_JFMC_CONF='$HOME/.jfrog/jfmc/jfmc-installer/jfmc-env.conf'
# JFMC_PORT can be set in ${JFMC_CONF} or mission-control.properties. The value in ${JFMC_CONF} takes priority 
notice "Using JFMC_PORT port : ${JFMC_PORT} (can be set or modified as \"export JFMC_PORT=<number>\" in ${DISPLAY_JFMC_CONF})\n"

# create all files (except for jfmc-conf) in data folder
INSTALLER_DATA_FOLDER="/jfmc_global_mount_root/jfmc-installer"
INSTALLED_VERSION_FILE_NAME="version.current"
INSTALLED_VERSION_FILE="${INSTALLER_DATA_FOLDER}/$INSTALLED_VERSION_FILE_NAME"
INSTALLED_VERSION_FILE_HISTORY="${INSTALLER_DATA_FOLDER}/version.history"

# Create install folder if its not avialable
mkdir -p "${INSTALLER_DATA_FOLDER}" || exitOnError "Could not create ${DISPLAY_INSTALL_LOG_LOCATION} directory, \nPlease run the following command on host to fix this mkdir -p ${DISPLAY_INSTALL_LOG_LOCATION}"

: ${INSTALL_ELASTIC_SEARCH:="Y"}
: ${INSTALL_POSTGRES:="Y"}

INSTALLED_VERSION=''
JFMC_GLOBAL_MOUNT_ROOT="/jfmc_global_mount_root" # This should not be used, unless special maintenance is needed
JFMC_SECURITY_DIR="${JFMC_GLOBAL_MOUNT_ROOT}/jfmc/data/security"
JFMC_MASTER_KEY_FILE_NAME="mc.key"
JFMC_MASTER_KEY_FILE_PATH="${JFMC_SECURITY_DIR}/${JFMC_MASTER_KEY_FILE_NAME}"

loadInstalledVersion(){
    if [ -f $INSTALLED_VERSION_FILE ]
    then
        INSTALLED_VERSION=$(cat $INSTALLED_VERSION_FILE)
    fi
}
loadInstalledVersion

# Maintainer / Installer version
MAINTAINER_VERSION_FILE="/opt/jfrog/jfmc/installer/version.current"
MAINTAINER_VERSION=''
if [ -f $MAINTAINER_VERSION_FILE ]
then
    MAINTAINER_VERSION=$(cat $MAINTAINER_VERSION_FILE)
fi

export IMAGE_MOUNT_ROOT="/var/opt/jfrog/mission-control"
export JFMC_PORT


# Utility method to remove environment variables
# 1st parameter "the key to remove"
removeFromEnvFile(){
    local key_to_remove=$1
    #Delete the original entry. NOTE: This will also ensure that if a value is not available, it is removed from the file
    sed -i "/export $key_to_remove/d" $JFMC_CONF
}

# Utility method to add environment variables
# 1st parameter "the key to add"
addToEnvFile(){
    if [ ! -f $JFMC_CONF ]; then
        return
    fi
    local key_to_add=$1
    # Get the value associated with this variable (indirect reference)
    local value_to_add=${!1}
    export $key_to_add="$value_to_add"

    removeFromEnvFile $key_to_add
    if [[ ! -z "$value_to_add" ]]; then
        #log "Executing echo export $key_to_add=\"$value_to_add\" >> $JFMC_CONF"
        echo "export $key_to_add=\"$value_to_add\"" >> ${JFMC_CONF}
    fi
}

#Given a file, this function runs the jq command and saves it back
replace_value() {
	file_name=$1
	jq_command=$2

	cat $file_name|jq $jq_command > ~/tmpfile && mv ~/tmpfile $file_name
}

usage() {
    cat ./readme.txt
    exit 1
}

version_lt() { test "$(echo "$@" | tr " " "\n" | sort -rV | head -n 1)" != "$1"; }

# $1 should be WARNING or ERROR
printUlimitErrorMsg()
{
    local error_level=$1
    local number=$2
    MSG="
${error_level}: Max number of open files [${number}] is too low !!!

Add the following to /etc/security/limits.conf file on HOST:
* soft nofile ${MAX_OPEN_FILES_REQUIRED_VALUE}
* hard nofile ${MAX_OPEN_FILES_REQUIRED_VALUE}

Note: Also verify docker limits configuration if exists [usually in /etc/default/docker]

"

    case ${error_level} in
    "ERROR")
        echo -e "\033[31m${MSG}\033[0m "
    ;;
    "WARNING")
        echo -e "\033[33m${MSG}\033[0m"
    ;;
    esac

    echo
}

checkPrerequisites()
{
    #echo "Check dependencies:"

    if [ -z ${DOCKER_SERVER_CURRENT_VER} ] || [ "${DOCKER_SERVER_CURRENT_VER}" == "" ]
    then
        echo "WARNING: Docker daemon version is unknown"
        # Do not exit on warning
    fi

    if version_lt ${DOCKER_SERVER_CURRENT_VER} ${DOCKER_SERVER_MIN_VER}
    then
        echo ""
        echo "ERROR: Mission Control requires Docker daemon version ${DOCKER_SERVER_MIN_VER} or later. The current version is ${DOCKER_SERVER_CURRENT_VER}."
        return 1
    fi

    local max_map_count=$(cat /proc/sys/vm/max_map_count)
    local recommended_map_count=262144
    if [ $max_map_count -lt $recommended_map_count ]; then
        exitOnError  "Not enough virtual memory to run $ELASTIC_SEARCH_LABEL reliably. Recommended: [vm.max_map_count=$recommended_map_count]. \
Please refer to [https://www.elastic.co/guide/en/elasticsearch/reference/current/docker.html#docker-cli-run-prod-mode]"
    fi


    # echo "Check max number of open files:"
    currentMaxOpenFilesVal=$(ulimit -n)

    if [ "${currentMaxOpenFilesVal}" != "unlimited" ] ; then

        # MAX_OPEN_FILES_REQUIRED_VALUE
        if [ ${currentMaxOpenFilesVal} -lt ${MAX_OPEN_FILES_REQUIRED_VALUE} ]
        then
            #echo "Trying to set max open files = ${MAX_OPEN_FILES_REQUIRED_VALUE}"
            ulimit -n ${MAX_OPEN_FILES_REQUIRED_VALUE} 2>/dev/null
            if [ $? -gt 0 ]
            then
                # Cannot set MAX_OPEN_FILES_REQUIRED_VALUE, will try to set MAX_OPEN_FILES_MIN_VALUE
                # MAX_OPEN_FILES_MIN_VALUE
                if [ ${currentMaxOpenFilesVal} -lt ${MAX_OPEN_FILES_MIN_VALUE} ]
                then
                    #echo "Trying to set max open files = ${MAX_OPEN_FILES_MIN_VALUE}"
                    ulimit -n ${MAX_OPEN_FILES_MIN_VALUE} 2>/dev/null
                    if [ $? -gt 0 ]; then
                        printUlimitErrorMsg "ERROR" "${currentMaxOpenFilesVal}"
                        return 1
                    fi
                fi
                printUlimitErrorMsg "WARNING" "${currentMaxOpenFilesVal}"
                # Do not exit on warning
            fi
        fi

    fi

    # dependencies check is passed
    return 0
}

runDockerCommand() {
    local action=$1

    if [[ $OFFLINE_MODE == true && $action == "pull" ]]; then
        return
    fi

    #If a modified docker-compose file has been created, use that instead
    if [ -f ${INSTALLER_DATA_FOLDER}/docker-compose.json ]; then
        cp ${INSTALLER_DATA_FOLDER}/docker-compose.json ./docker-compose.json
    fi

    case $action in
    "pull")
        if [ ! -z "$JFMC_DOCKER_REPO_LOGIN" ]; then
            docker login -u $JFMC_DOCKER_REPO_LOGIN -p $JFMC_DOCKER_REPO_PASSWORD $JFMC_DOCKER_REPO
        fi
    ;;
    "logs")
        action="logs -f"
    ;;
    "stop")
        # Stop Mission Control App layer first, than the DB layer
        docker-compose -f docker-compose.json -p ${COMPOSE_PROJECT_NAME} ${action} ${JFMC_APP_SERVICES_LIST}
    ;;
    "stopApp")
        # Stop Mission Control App layer first, than the DB layer
        docker-compose -f docker-compose.json -p ${COMPOSE_PROJECT_NAME} "stop" ${JFMC_APP_SERVICES_LIST}
        return $?
    ;;
    esac

    docker-compose -f docker-compose.json -p ${COMPOSE_PROJECT_NAME} ${action}
}

# Returned code:
#0) $1=$2
#1) $1>$2
#2) $1<$2
verComp () {
    if [[ $1 == $2 ]]
    then
        return 0
    fi
    local IFS=.
    local i
    local ver1=($1)
    local ver2=($2)
    # fill empty fields in ver1 with zeros
    for ((i=${#ver1[@]}; i<${#ver2[@]}; i++))
    do
        ver1[i]=0
    done
    for ((i=0; i<${#ver1[@]}; i++))
    do
        if [[ -z ${ver2[i]} ]]
        then
            # fill empty fields in ver2 with zeros
            ver2[i]=0
        fi
        if ((10#${ver1[i]} > 10#${ver2[i]}))
        then
            return 1
        fi
        if ((10#${ver1[i]} < 10#${ver2[i]}))
        then
            return 2
        fi
    done
    return 0
}

copyFileNoOverwrite() {
    source=$1
    target=$2
    
    [[ ! -z ${source} ]] || exitOnError "Source file to copy not provided"
    [[ ! -z ${target} ]] || exitOnError "Target file for copy not provided"

    if ! [[ -f ${target} ]]; then
        cp -f ${source} ${target} || exitOnError "Copy of file from ${source} to ${target} failed, make sure parent of ${target} location has the right permissions"
    fi
}

isUpgradeAllowed() {

    local installedVersion=$INSTALLED_VERSION
    local maintainerVersion=$MAINTAINER_VERSION

    if [ -z $MAINTAINER_VERSION ]
    then
        exitOnError "MAINTAINER_VERSION is unknown"
    fi

    # This is a fresh installation
    if [ -z $INSTALLED_VERSION ]
    then
        return 0
    fi

    if [[ "$installedVersion" =~ ^b.* ]] && [[ "$maintainerVersion" =~ ^b.* ]]
    then
        installedVersion=($(echo $INSTALLED_VERSION | sed "s/^b//g"))
        maintainerVersion=($(echo $MAINTAINER_VERSION | sed "s/^b//g"))
    fi

    verComp $installedVersion $maintainerVersion 2>/dev/null
    if [ $? -eq 1 ]
    then
        return 1
    fi

    return 0
}

isCommandAllowed()
{
    local command=$1
    local installedVersion=$INSTALLED_VERSION
    local maintainerVersion=$MAINTAINER_VERSION

    case ${command} in
        "install"|"installLocal")
            [ -f $INSTALLED_VERSION_FILE ] && exitOnError "Mission Control is already installed. To upgrade, use: ./mission-control upgrade"
            checkPrerequisites || exitOnError "Prerequisites are not satisfied"
        ;;
        "upgrade")
            [ -f $INSTALLED_VERSION_FILE ] || exitOnError "Mission Control is not installed. To install, use: ./mission-control install"
            checkPrerequisites || exitOnError "Prerequisites are not satisfied"
            isUpgradeAllowed || exitOnError "Unable to upgrade to an older version"
        ;;
        "version")

        ;;
        "info")

        ;;
        *)
            # Is installed
            [ -f $INSTALLED_VERSION_FILE ] || exitOnError "Mission Control is not installed. To install, use: ./mission-control install"
            # Is correct version
            [ "${installedVersion}" == "${maintainerVersion}" ] || exitOnError "You are using the Mission Control installer version [${maintainerVersion}] which is different from the currently installed version [${installedVersion}]. Consider upgrading your version of Mission Control using: ./mission-control upgrade"
        ;;

    esac

    return 0
}

setVersions()
{
    if [ ! -f ${MAINTAINER_VERSION_FILE} ]
    then
        return 1
    fi

    cat ${MAINTAINER_VERSION_FILE} > ${INSTALLED_VERSION_FILE}
    cat ${MAINTAINER_VERSION_FILE} >> ${INSTALLED_VERSION_FILE_HISTORY}

    echo "${INSTALLED_VERSION_FILE}: $(cat ${INSTALLED_VERSION_FILE})"

    return 0
}

createPostgresUsers(){
    docker cp postgres "${COMPOSE_PROJECT_NAME}_postgres_1":/ && \
    docker cp $JFMC_CONF_DEFAULT "${COMPOSE_PROJECT_NAME}_postgres_1":/postgres/setenvDefaults.sh && \
    ( [ -f $JFMC_CONF ] && docker cp $JFMC_CONF "${COMPOSE_PROJECT_NAME}_postgres_1":/postgres/setenv.sh || true ) && \
    docker exec -i "${COMPOSE_PROJECT_NAME}_postgres_1" bash -c "gosu postgres /postgres/createPostgresUsers.sh" || \
    exitOnError "${POSTGRES_LABEL} setup failed" 

    # Cleanup setup files from postgres container
    docker exec -i "${COMPOSE_PROJECT_NAME}_postgres_1" bash -c "rm -r /postgres" ;
}

isHAInstall() {
    if [[ $JFMC_HA_INSTALLATION == true ]]; then
        return 0
    else
        return 1
    fi
}

isESInstalled() {
    if [[ $INSTALL_ELASTIC_SEARCH =~ y|Y ]]; then
        return 0
    else
        return 1
    fi
}

externaliseES() {
    if [[ $INSTALL_ELASTIC_SEARCH =~ n|N ]]; then
        return 0
    else
        return 1
    fi
}

isNotFirstNode () {
    if [[ $JFMC_FIRST_NODE =~ n|N ]]; then
        return 0
    else
        return 1
    fi
}

setEnv() {
    local key=$1
    local value=$2

    if [ -z "${key}" ] || [ -z "${key}" ]; then
        return
    fi

    export ${key}=${value}
    addToEnvFile "${key}"
}

setESEnv() {
    if ! isESInstalled || ! isHAInstall; then
        return
    fi

    setHostIP
    setEnv "JFMC_ES_CLUSTER_SETUP" "YES"

    if isNotFirstNode; then
        setEnv "ELASTIC_MIN_MASTER_NODES" 2
    fi
}

configureElasticSearch(){
    if isNotFirstNode; then
        log "Skipping initialisation of search guard, this will already be intialised in first node (JFMC_FIRST_NODE : ${JFMC_FIRST_NODE}) "
    else
        docker cp configureElasticsearch.sh "${COMPOSE_PROJECT_NAME}_elasticsearch_1":/usr/share/elasticsearch/ && \
        docker exec -i "${COMPOSE_PROJECT_NAME}_elasticsearch_1" bash -c "/usr/share/elasticsearch/configureElasticsearch.sh" || \
        exitOnError "${ELASTIC_SEARCH_LABEL} setup failed"
    fi
}

configureDatabases(){
    [[ $INSTALL_POSTGRES =~ y|Y ]] && createPostgresUsers || true
    isESInstalled && configureElasticSearch || true
}

executeMigrationIfNeeded()
{
    local migrationName=$1
    local migrationForVersion=$2
    verComp $INSTALLED_VERSION $migrationForVersion 2>/dev/null
    if [ $? -eq 2 ]
    then
        log "\nINFO: Executing migration from [${INSTALLED_VERSION}] to [${migrationForVersion}] - ${migrationName}"
        eval ${migrationName}
    fi
}

# Handle upgrade from v2.x to v3.4.x and above. This will reset old stock setup of postgres data with new setup.
moveOldPostgresData(){
    local postgresData="${JFMC_GLOBAL_MOUNT_ROOT}/postgres"
    local postgresDataV2x="${JFMC_GLOBAL_MOUNT_ROOT}/postgres_v2x"
    if [[ ! -z ${INSTALLED_VERSION} && ${INSTALLED_VERSION} == "2."* && -d ${postgresData} ]]; then
        moveItem "${postgresData}" "${postgresDataV2x}"
    fi
}

moveItem () {
    local source=$1
    local target=$2

    [ ! -z ${source} ] || return
    [ ! -z ${target} ] || return

    log "INFO: Moving ${source} to ${target}"
    mv "${source}" "${target}" || true
}

# Migrate old mission-control logs from logs root directory to service specifice directory
migrateMCLogsToNewStructure()
{
    local logsRoot="$1"
    local jfmcAppName=${JFMC_APP_NAME:-"jfmc-server"}

    [[ ! -z ${logsRoot} ]] || return
    [[ -d ${logsRoot}   ]] || return

    local mcLogs=${logsRoot}/${jfmcAppName}
    local oldAppLog="${logsRoot}/mission-control.log"
    local newAppLog="${mcLogs}/${jfmcAppName}.log"

    local oldArchive="${logsRoot}/archived"
    local newArchive="${mcLogs}"/

    # Create new logs directory if its not available
    [[ -d ${mcLogs} ]] ||  ( mkdir -p ${mcLogs} || return )

    if [ -d "${oldArchive}" ]; then
        moveItem "${oldArchive}" "${newArchive}"
    fi

    if [ -f "${oldAppLog}" ]; then
        moveItem "${oldAppLog}" "${newAppLog}"
    fi

    for fileToMigrate in "monitoring.log" "audit.log"
    do
        local oldFile="${logsRoot}/${fileToMigrate}"
        local newFile="${mcLogs}/${fileToMigrate}"

        if [ -f "${oldFile}" ]; then
            moveItem "${oldFile}" "${newFile}"
        fi
    done
}

# backward compatible issues
# Mission Control should be stopped before this process
backwardCmpatibleProcess()
{
    log "\nINFO: Executing possible migration steps"
    moveOldPostgresData
    executeMigrationIfNeeded "migrateToSingleMount" "2"
    migrateMCLogsToNewStructure "${JFMC_GLOBAL_MOUNT_ROOT}/jfmc/logs"
}

migrateToSingleMount()
{
    local JFMC_MOUNT_V1="${JFMC_GLOBAL_MOUNT_ROOT}"
    local JFMC_MOUNT_V2="${JFMC_GLOBAL_MOUNT_ROOT}/jfmc"
    log "\nINFO: Moving data,etc,logs from [${JFMC_MOUNT_V1}] to [${JFMC_MOUNT_V2}]"

    if [ ! -d ${JFMC_MOUNT_V2} ]; then
        mkdir -p ${JFMC_MOUNT_V2} && \
        echo "\nINFO: New folder at [${JFMC_MOUNT_V2}] was created"
    fi

    for folderToMigrate in "logs" "etc" "data"
    do
        local folderToMigrateV1="${JFMC_MOUNT_V1}/${folderToMigrate}"
        local folderToMigrateV2="${JFMC_MOUNT_V2}/${folderToMigrate}"
        if [ -d "${folderToMigrateV1}" ] && [ ! -d "${folderToMigrateV2}" ]
        then
            mv ${folderToMigrateV1} ${folderToMigrateV2} && \
            log "\nINFO: ${folderToMigrate} folder renamed was:[${folderToMigrateV1}] new:[${folderToMigrateV2}]"
        fi
    done
}

# Utility method to check if a value (first parameter) exists in an array (2nd parameter)
# 1st parameter "value to find"
# 2nd parameter "The array to search in. Please pass a string with each value separated by a comma"
# Example: containsElement "y" "y Y n N"
containsElement () {
  local searchElement=$1
  local searchArray=($2)
  local found=1
  for elementInIndex in "${searchArray[@]}";do
    if [[ $elementInIndex == $searchElement ]]; then
        found=0
    fi
  done
  return $found
}

# Utility method to get user's choice
# 1st parameter "what to ask the user"
# 2nd parameter "what choices to accept, separated by spaces"
# 3rd parameter "what is the default choice (to use if the user simply presses Enter)"
# Example 'getUserChoice "Are you feeling lucky? Punk!" "y n Y N" "y"'

getUserChoice(){
    local choice="na"
    local text_to_display=$1
    local choices=$2
    local default_choice=$3

    until containsElement "$choice" "$choices"; do
        echo;
        sleep $read_timeout #This ensures correct placement of the question.
        if [[ $USE_DEFAULTS == "true" ]]; then
            choice=$default_choice
        else
            read -p "$text_to_display:" choice
            : ${choice:=$default_choice}
        fi
    done
    users_choice=$choice
    if [[ -f $INSTALLATION_LOG_FILE ]]; then
        #Ensures the choices made by the user are also put into the log file
        echo "$text_to_display: $users_choice" >> $INSTALLATION_LOG_FILE
    fi
}

# Utility method to get user's input
# 1st parameter "what to ask the user"
# 3rd parameter "what is the default choice (to use if the user simply presses Enter)"
# Example 'getUserInput "Are you feeling lucky? Punk!" "y"'
getUserInput(){
    local choice="na"
    local text_to_display=$1
    local default_choice=$2
    users_choice=
    
    if [ ! -z $default_choice ]; then
        text_to_display="$1 [$default_choice]:"
    else
        text_to_display="$1 :"
    fi

    echo "";echo "";
    sleep $read_timeout #This ensures correct placement of the question.
    read -p "$text_to_display" choice
    if [ ! -z $default_choice ]; then
        : ${choice:=$default_choice}
    fi
    users_choice=$choice
    sleep $read_timeout #This ensures correct logging
}

getElasticSearchDetails() {
    #Variables for Elastic search
    ELASTIC_SEARCH_URL="http://host.docker.internal:9200"
    #Ask the user the URL to connect to
    
    getUserInput "Please enter the $ELASTIC_SEARCH_LABEL URL" "$ELASTIC_SEARCH_URL"
    ELASTIC_SEARCH_URL=$users_choice
    ELASTIC_SEARCH_WRITE_URL=$users_choice
    ELASTIC_SEARCH_READ_URL=$users_choice

    addToEnvFile "ELASTIC_SEARCH_URL"
    addToEnvFile "ELASTIC_SEARCH_WRITE_URL"
    addToEnvFile "ELASTIC_SEARCH_READ_URL"

    warn "If this $ELASTIC_SEARCH_LABEL instance needs credentials, please update the values corresponding to 'elastic.username' and 'elastic.password' in the file: ${JFMC_MOUNT_ROOT}/jfmc/etc/mission-control.properties before starting the application. Please create the file with these values if it is not available."
}

externalizeElasticSearch() {
    title "$ELASTIC_SEARCH_LABEL"
    log "Install $ELASTIC_SEARCH_LABEL as part of this installation? \
(Choose 'N' if you want to externalize $ELASTIC_SEARCH_LABEL)"
    getUserChoice "Install $ELASTIC_SEARCH_LABEL? [Y/n]" "y n Y N" "y"

    INSTALL_ELASTIC_SEARCH=${users_choice}
    addToEnvFile "INSTALL_ELASTIC_SEARCH"
    
    if externaliseES; then
        getElasticSearchDetails
        updateElasticSearchDetails
    else
        removeFromEnvFile "ELASTIC_SEARCH_URL"
        removeFromEnvFile "ELASTIC_SEARCH_WRITE_URL"
        removeFromEnvFile "ELASTIC_SEARCH_READ_URL"
        loadDefaultElasticSearchEnv
    fi
}

updateElasticSearchDetails() {
    local tmpFile=~/"tmpfile"
    local dockerCompose="docker-compose.json"

    if externaliseES; then
        #Remove the elastic search container from the docker-compose file
        cat ${dockerCompose}|jq 'del(.services.elasticsearch)' > ${tmpFile} && mv ${tmpFile} ${dockerCompose}
        for service in insight_server ${OPTIONAL_SERVICES}; do
            #Remove the entry that makes services depend on elasticsearch container
            cat ${dockerCompose}|jq 'del(.services.'${service}'.depends_on[] | select( . == "elasticsearch") )' > ${tmpFile} && mv ${tmpFile} ${dockerCompose}
        done
    fi
        
}

# Add an entry to an array in docker-compose.json
#   service    : service to update
#   arrayPath  : path to array
#   arrayValue : value to be added
appendJSONArrayItem(){
    local service=$1
    local arrayPath=$2
    local arrayValue=$3
    local tmpFile=~/"tmpfile"
    local dockerCompose="docker-compose.json"

    cat ${dockerCompose} | grep -wq '"'${service}'"' ; EXIT_CODE=$?
    if [[ -z $EXIT_CODE || $EXIT_CODE == 0 ]]; then
        #Remove the entry if its available - to avoid duplicate entry
        cat ${dockerCompose}|jq 'del('${arrayPath}'[] | select( . == "'${arrayValue}'") )' > ${tmpFile} ; mv ${tmpFile} ${dockerCompose}
        #Add the new entry
        cat ${dockerCompose}|jq ${arrayPath}' += [ "'${arrayValue}'" ]' > ${tmpFile} && mv ${tmpFile} ${dockerCompose}
    fi
}

# Create a unicast file which can be accessed by insight_Server and elastisearch
# Insight server will write elasticsearch HA cluster details
# Elasticsearch will read this to discover master nodes
setUpUnicastFile() {
    local elasticsearchConfDir="elasticsearch/config"
    local elasticsearchConf="${JFMC_GLOBAL_MOUNT_ROOT}/${elasticsearchConfDir}"
    local elasticsearchUnicastFileName="unicast_hosts.txt"
    local elasticsearchUnicast="${elasticsearchConf}"/${elasticsearchUnicastFileName}
    local esGID=1000

    mkdir -p ${elasticsearchConf} || exitOnError "Creating mount directories failed : ${JFMC_MOUNT_ROOT}/${elasticsearchConfDir} \nPlease run the following command on host to fix this : mkdir -p ${JFMC_MOUNT_ROOT}/${elasticsearchConfDir} && chown -R ${esGID}:${esGID} ${JFMC_MOUNT_ROOT}/${elasticsearchConfDir}"

    # Create unicast file with ownership 664
    # Insight server microservice will write node entries and elasticsearch will read them through this file
    [[ -f "${elasticsearchUnicast}" ]]  || echo "" > ${elasticsearchUnicast}  || exitOnError "Creating an empty file failed : ${JFMC_MOUNT_ROOT}/${elasticsearchConfDir}/${elasticsearchUnicastFileName} \nPlease run the following command on host to fix this : echo "" > ${JFMC_MOUNT_ROOT}/${elasticsearchConfDir}/${elasticsearchUnicastFileName}"
    chmod 664 "${elasticsearchUnicast}" || exitOnError "Changing the file mod failed : ${JFMC_MOUNT_ROOT}/${elasticsearchConfDir}/${elasticsearchUnicastFileName} \nPlease run the following command on host to fix this : chmod 664 ${JFMC_MOUNT_ROOT}/${elasticsearchConfDir}/${elasticsearchUnicastFileName}"
}

setupESHAMount(){
    if ! isESInstalled || ! isHAInstall; then
        return
    fi

    #Create unicast file in the mount directoy and its mount entries in docker-compose.json
    setUpUnicastFile
    appendJSONArrayItem "insight_server" ".services.insight_server.volumes" '${JFMC_MOUNT_ROOT}/elasticsearch/config/unicast_hosts.txt:${JFIS_ELASTIC_UNICAST_HOST_FILE}'
    appendJSONArrayItem "elasticsearch"  ".services.elasticsearch.volumes"  '${JFMC_MOUNT_ROOT}/elasticsearch/config/unicast_hosts.txt:/usr/share/elasticsearch/config/unicast_hosts.txt'
}

setClusterConfig() {
    if [ ${USE_DEFAULTS} == "true" ] ; then
        # Force standalone installation in case of "use defaults" mode and no input from DISPLAY_JFMC_CONF
        JFMC_HA_INSTALLATION=false
    else
        getUserChoice "Is this a HA (High Availability) installation [y/N] " "y n Y N" "N"
        if [[ "${users_choice}" = [yY] ]]; then
        
            JFMC_HA_INSTALLATION=true
            getUserChoice "Are you adding this node to an existing cluster? (not relevant for the first cluster node) [y/N] " "y n Y N" "N"
            
            if [[ "${users_choice}" = [yY] ]]; then
                local jfmcMasterKey=
                until [[ ! -z ${jfmcMasterKey} ]]; do
                    echo;
                    sleep .5;
                    read -p "Provide the master key from the first node in cluster (the key can be found in <JFMC DATA FOLDER>/security/${JFMC_MASTER_KEY_FILE_NAME}) : " jfmcMasterKey
                done

                mkdir -p "${JFMC_SECURITY_DIR}" || exitOnError "Failed to create $JFMC_MOUNT_ROOT/jfmc/data/security directory, please make sure $USER has permission to perform this action"
                echo -n "${jfmcMasterKey}" > "${JFMC_MASTER_KEY_FILE_PATH}" || exitOnError "Failed to write master key into $JFMC_MOUNT_ROOT/jfmc/data/security/${JFMC_MASTER_KEY_FILE_NAME}, please make sure $USER has permission to perform this action"
                JFMC_FIRST_NODE="N"
            else
                notice "A master key will generated after start up in ${JFMC_MOUNT_ROOT}/jfmc/data/security/${JFMC_MASTER_KEY_FILE_NAME}, installer will ask for this key during the installation of other nodes."
                JFMC_FIRST_NODE="Y"
            fi
        else
            JFMC_HA_INSTALLATION=false
        fi
    fi

    if isHAInstall ; then
        NODEHEALTHCHECK_KILL_ONMAXFAILURES=true
    fi

    addToEnvFile "NODEHEALTHCHECK_KILL_ONMAXFAILURES"
    addToEnvFile "JFMC_HA_INSTALLATION"
    addToEnvFile "JFMC_FIRST_NODE"
}

# Add mongo service and connection details based on INSTALL_MONGO env variable
# This needs to come up to migrate data from mongo to postgres
# Handles :
#       Standard upgrade with stock mongo
#       Standard upgrade with externalised mongo
# Can be disabled for next upgrade by setting the following environment variable in ${DISPLAY_JFMC_CONF},
#       export INSTALL_MONGO=NOT_APPLICABLE
addMongoMigrationDetails() {
    local tmpFile=/opt/"tmpfile"
    local dockerCompose=${INSTALLER_DATA_FOLDER}/"docker-compose.json"
    local dockerComposeBackup=${INSTALLER_DATA_FOLDER}/"docker-compose.json.backup"
    local mongoCompose="mongo-compose.json"
    local mongoVariables="mongo-variables.json"
    local wasMongoInstalled=

    # Check if previous install had mongodb
    if [ -f ${dockerComposeBackup} ]; then 
        cat ${dockerComposeBackup} | grep -wq '"'mongodb'"' ; wasMongoInstalled=$?
    fi

    # Set INSTALL_MONGO to NA or NOT_APPLICABLE or DEPRECATED to avoid starting mongodb on next upgrade
    if [[ $INSTALL_MONGO =~ (n|N|DEP)+ ]]; then
        return;
    elif [[ $INSTALL_MONGO =~ y|Y ]] || [[ $wasMongoInstalled == 0 ]] || [[ "${INSTALLED_VERSION}" == "2.0.0" ]] ; then
        notice "Adding mongodb service to compose for migration of data from mongodb to postgres \n Add the following entry \"export INSTALL_MONGO=NOT_APPLICABLE\" in ${DISPLAY_JFMC_CONF} to remove the container in the next upgrade"
        # Add mongo as a dependency to start it before other services
        cat ${dockerCompose} | jq '.services.server.depends_on += [ "mongodb" ]' > ${tmpFile} && mv ${tmpFile} ${dockerCompose}
        # Add mongo service node to compose and pass relevant mongo environment variables
        jq -s '.[0] * .[1]' ${dockerCompose} ${mongoCompose}   > ${tmpFile} && mv ${tmpFile} ${dockerCompose} || ( warn "Unable to add mongo service to the main compose, this is needed for migration. \n If migration is complete, remove INSTALL_MONGO variable from ${JFMC_CONF} or set it to NOT_APPLICABLE and execute ./mission-control upgrade" && return )
        jq -s '.[0] * .[1]' ${dockerCompose} ${mongoVariables} > ${tmpFile} && mv ${tmpFile} ${dockerCompose} || ( warn "Unable to add mongo variables to the main compose, this is needed for migration. \n If migration is complete, remove INSTALL_MONGO variable from ${JFMC_CONF} or set it to NOT_APPLICABLE and execute ./mission-control upgrade" && return )
    elif [[ $INSTALL_MONGO =~ n|N ]] && [[ $MONGODB_HOST != "mongodb" ]]; then
        notice "Adding mongodb connection details to compose for migration of data from mongodb to postgres \n Add the following entry \"export INSTALL_MONGO=NOT_APPLICABLE\" in ${DISPLAY_JFMC_CONF} to remove the mongo connection details from compose in the next upgrade"
        # Add mongo as a dependency to start it before other services
        cat ${dockerCompose} | jq '.services.server.depends_on += [ "mongodb" ]' > ${tmpFile} && mv ${tmpFile} ${dockerCompose}
        # Pass externalised mongodb details to all the microservices
        jq -s '.[0] * .[1]' ${dockerCompose} ${mongoVariables} > ${tmpFile} && mv ${tmpFile} ${dockerCompose} || ( warn "Unable to add mongo variables to the main compose, this is needed for migration. \n If migration is complete, remove INSTALL_MONGO variable from ${JFMC_CONF} or set it to NOT_APPLICABLE and execute ./mission-control upgrade" && return )
    fi
}

externalizePostgres() {
    title "$POSTGRES_LABEL"
    if isHAInstall; then
        INSTALL_POSTGRES=N
        addToEnvFile "INSTALL_POSTGRES"

        getPostgresDetails
        seedPostgres
        updatePostgresDetails
    else
        log "Install $POSTGRES_LABEL as part of this installation? (Choose 'n' if you want to externalize $POSTGRES_LABEL)"
        getUserChoice "Install $POSTGRES_LABEL? [Y/n]" "y n Y N" "y"
        INSTALL_POSTGRES=${users_choice}
        addToEnvFile "INSTALL_POSTGRES"
        if [[ $INSTALL_POSTGRES =~ n|N ]]; then
            getPostgresDetails
            seedPostgres
            updatePostgresDetails
        else
            removeFromEnvFile "DB_HOST"
            removeFromEnvFile "DB_PORT"
            loadDefaultPostgresEnv
        fi
    fi
}

getPostgresDetails() {
    #Variables for Postgres
    DB_HOST="host.docker.internal"
    : ${DB_PORT:="5432"}

    #Ask the host to connect to
    getUserInput "Please enter the $POSTGRES_LABEL Host" "$DB_HOST"
    DB_HOST=$users_choice
    addToEnvFile "DB_HOST"

    #Ask the port to connect to
    getUserInput "Please enter the $POSTGRES_LABEL Port" "$DB_PORT"
    DB_PORT=$users_choice
    addToEnvFile "DB_PORT"
}

updatePostgresDetails() {
    local tmpFile=/opt/"tmpfile"
    local dockerCompose="docker-compose.json"
    
    if [[ $INSTALL_POSTGRES =~ n|N ]]; then
        #Remove the postgres container from the docker-compose file
        cat ${dockerCompose}|jq 'del(.services.postgres)' > ${tmpFile} && mv ${tmpFile} ${dockerCompose}
        cat ${dockerCompose}|jq 'del( .services.server.links[] | select( . == "postgres") )'  > ${tmpFile} && mv ${tmpFile} ${dockerCompose}

        for service in server insight_server scheduler executor; do
            #Remove the entry that makes services depend on postgres container
            cat ${dockerCompose}|jq 'del( .services.'${service}'.depends_on[] | select( . == "postgres") )'  > ${tmpFile} && mv ${tmpFile} ${dockerCompose}
        done
    fi
}

copySeedData() {
    mkdir -p $INSTALLER_DATA_FOLDER/postgres
    cp postgres/* $INSTALLER_DATA_FOLDER/postgres
}

seedPostgres() {
    warn "Files which can be used to seed $POSTGRES_LABEL are now available at [$JFMC_MOUNT_ROOT/jfmc-installer/postgres] \
Please seed $POSTGRES_LABEL before starting Mission-Control" #Don't add a tab to this line
}

MAX_WAIT_TIME=120
performHealthCheck() {
    local URL="${JFMC_URL}/api/v3/system/health_check/node"
    local expectedOutput=0

    log "Healthcheck in progress using [$URL]"
    local actualOutput=1
    local curl_response
    # wait for service to be up, check every 2 secs
    # wait for a maximum of 50 seconds, so it doesn't get into an infinite loop
    WAIT_TIME=0
    while [[ $actualOutput -ne $expectedOutput ]] && [[ $WAIT_TIME -lt $MAX_WAIT_TIME ]] ; do
            echo -n "."
            sleep 2
            WAIT_TIME=$((WAIT_TIME + 3))
            curl -s --max-time 30 -I -X GET "$URL" | grep "HTTP.*\s2[0-9]*" >/dev/null 2>/dev/null
            actualOutput=$?
    done
    if [[ $actualOutput -eq $expectedOutput ]]
    then
        log "Status: OK"
    else
        log "Healthcheck Response:"
        curl -X GET "$URL"
        exitOnError "Mission control is up but reporting issues. Please inspect the logs and take necessary steps to fix the problem.\nNote : The health check call from installer can fail if SSL is enabled on mission-control with a self signed certificate (can be ignored if the application is running fine)."
    fi
}

connectNetwork(){
    docker network connect jfmc_default jfmc_installer
}

healthCheck(){
    connectNetwork
    performHealthCheck 
}

backupComposeFile() {
    if [ -f ${INSTALLER_DATA_FOLDER}/docker-compose.json ]; then
        cp ${INSTALLER_DATA_FOLDER}/docker-compose.json ${INSTALLER_DATA_FOLDER}/docker-compose.json.backup
        warn "The current docker-compose file has been backed up as docker-compose.json.backup. \n\
Once the upgrade is complete, you can reconcile any manual changes you may have made to the file"
    else
        cp docker-compose.json ${INSTALLER_DATA_FOLDER}/docker-compose.json
    fi
}

chooseTypeOfUpgrade(){
    log "Would you prefer a standard Upgrade? (Choose 'N' if you want to change any choices you made earlier)"

    getUserChoice "Perform a standard Upgrade? [Y/n]" "y n Y N" "y"
    STANDARD_UPGRADE=${users_choice}

    source "${JFMC_CONF}"
    if [[ $STANDARD_UPGRADE =~ y|Y ]]; then
        updatePostgresDetails
        updateElasticSearchDetails
    else
        warn "$LOCALHOST_WARNING"
        externalizePostgres
        externalizeElasticSearch
    fi
    cp docker-compose.json ${INSTALLER_DATA_FOLDER}/docker-compose.json
}

chooseTypeOfInstallation(){
    if isHAInstall; then
        STANDARD_INSTALLATION=N
    else
        title "Choose Type of Installation"
        log "Mission Control uses $ELASTIC_SEARCH_LABEL and $POSTGRES_LABEL. The installer can install these automatically. \n\
    Would you prefer a standard installation? (If you choose 'N', the installer will ask you before installing any 3rd-party database)"

        getUserChoice "Perform a standard Installation? [Y/n]" "y n Y N" "y"
        STANDARD_INSTALLATION=${users_choice}
    fi

    addToEnvFile "STANDARD_INSTALLATION"
    if [[ $STANDARD_INSTALLATION =~ n|N ]] || [[ $JFMC_HA_INSTALLATION == true ]]; then
        warn "$LOCALHOST_WARNING"
        externalizePostgres
        externalizeElasticSearch
    fi
    cp docker-compose.json ${INSTALLER_DATA_FOLDER}/docker-compose.json
}

# Move all installer files from user's home to data location in case of upgrades
# Keep only data location info in home location
migrateInstallInfo()
{
    if [[ ! -f "$INSTALLED_VERSION_FILE" && -f "$DEFAULT_INSTALLER_DATA_FOLDER/$INSTALLED_VERSION_FILE_NAME" ]]; then
        runDockerCommand stop \
        && mkdir -p "${INSTALLER_DATA_FOLDER}" \
        && cp -fr "$DEFAULT_INSTALLER_DATA_FOLDER"/* "${INSTALLER_DATA_FOLDER}"/ \
        && rm -fr "$DEFAULT_INSTALLER_DATA_FOLDER"/* \
        && cp "${INSTALLER_DATA_FOLDER}/$JFMC_CONF_FILE_NAME"  "${JFMC_CONF}" \
        && rm -f "${INSTALLER_DATA_FOLDER}/$JFMC_CONF_FILE_NAME" \
        && loadInstalledVersion \
        || exitOnError "Unable to migrate install information"
    fi
}

# Remove any ports exposed on host, reset exposed ports and add 8080 as the only exposed port    
removeExposedPorts() {
    #Don't remove exposed ports on dev machines
    if [ ! -z "$JFMC_DOCKER_REPO_LOGIN" ]; then
        return 0
    fi
    
    local tmpfile=/opt/tmpfile
    local dockerCompose="docker-compose.json"

    for service in elasticsearch postgres executor scheduler executor insight_server; do
        cat ${dockerCompose} | grep -wq '"'$service'"' ; EXIT_CODE=$?
        if [[ -z $EXIT_CODE || $EXIT_CODE == 0 ]]; then
            #log "Removing exposed ports on $ELASTIC_SEARCH_LABEL"
            cat ${dockerCompose}|jq 'del( .services.'${service}'.ports[] | select( . ) )'  > ${tmpfile} && mv ${tmpfile} ${dockerCompose}
            cat ${dockerCompose}|jq 'del( .services.'${service}'.expose[] | select( . ) )' > ${tmpfile} && mv ${tmpfile} ${dockerCompose}

            if [[ $service == "elasticsearch" ]]; then
                cat ${dockerCompose}|jq '.services.'${service}'.expose += [ "9200", "${ELASTIC_TRANSPORT_PORT}" ]' > ${tmpfile} && mv ${tmpfile} ${dockerCompose}
                if isHAInstall ; then
                    cat ${dockerCompose}|jq '.services.'${service}'.ports += [ "'${ELASTIC_TRANSPORT_PORT}':'${ELASTIC_TRANSPORT_PORT}'" ]' > ${tmpfile} && mv ${tmpfile} ${dockerCompose}
                fi
            elif [[ $service == "postgres" ]]; then
                cat ${dockerCompose}|jq '.services.'${service}'.expose += [ "5432" ]' > ${tmpfile} && mv ${tmpfile} ${dockerCompose}
            elif [[ $service == "scheduler" ]]; then
                cat ${dockerCompose}|jq '.services.'${service}'.expose += [ "'${JFMC_SCHEDULER_PORT}'" ]' > ${tmpfile} && mv ${tmpfile} ${dockerCompose}
            elif [[ $service == "executor" ]]; then
                cat ${dockerCompose}|jq '.services.'${service}'.expose += [ "'${JFMC_EXECUTOR_PORT}'" ]' > ${tmpfile} && mv ${tmpfile} ${dockerCompose}
            elif [[ $service == "insight_server" ]]; then
                cat ${dockerCompose}|jq '.services.'${service}'.expose += [ "'${JFMC_INSIGHT_SERVER_PORT}'" ]' > ${tmpfile} && mv ${tmpfile} ${dockerCompose}
            fi
        fi
    done
}

# Creating JFMC config dir and setting jfmc user ownership on files
setupDirsPermissions () {
    local installer_mount_root="${INSTALLER_DATA_FOLDER}"
    local installer="jfmc-installer"
    local jfmc="jfmc"
    local jfmc_mount_root="${JFMC_GLOBAL_MOUNT_ROOT}/${jfmc}"

    if isESInstalled; then
        local esGID=1000
        local elasticsearchRoot="elasticsearch"
        local elasticsearchData="${elasticsearchRoot}/data"
        local elasticsearchSGConf="${elasticsearchRoot}/sgconfig"
        local elasticsearchDataMountRoot="${JFMC_GLOBAL_MOUNT_ROOT}/${elasticsearchData}"
        local elasticsearchSGConfMountRoot="${JFMC_GLOBAL_MOUNT_ROOT}/${elasticsearchSGConf}"

        log "Setting ownership on ${JFMC_MOUNT_ROOT}/${elasticsearchRoot} on host"
        mkdir -p ${elasticsearchDataMountRoot} ${elasticsearchSGConfMountRoot}                   || exitOnError "Creating mount directories failed : ${JFMC_MOUNT_ROOT}/${elasticsearchData} ${JFMC_MOUNT_ROOT}/${elasticsearchSGConf} \nPlease run the following command on host to fix this : mkdir -p ${JFMC_MOUNT_ROOT}/${elasticsearchData} ${JFMC_MOUNT_ROOT}/${elasticsearchSGConf} && chown -R ${esGID}:${esGID} ${JFMC_MOUNT_ROOT}/${elasticsearchData} ${JFMC_MOUNT_ROOT}/${elasticsearchSGConf}"
        chown -R ${esGID}:${esGID} ${elasticsearchDataMountRoot} ${elasticsearchSGConfMountRoot} || exitOnError "Setting ownership on following directories failed : ${JFMC_MOUNT_ROOT}/${elasticsearchData} ${JFMC_MOUNT_ROOT}/${elasticsearchSGConf} \nPlease run the following command on host to fix this : chown -R ${esGID}:${esGID} ${JFMC_MOUNT_ROOT}/${elasticsearchData} ${JFMC_MOUNT_ROOT}/${elasticsearchSGConf}"
    fi

    mkdir -p ${installer_mount_root} ${jfmc_mount_root} || exitOnError "Creating mount directories failed : ${JFMC_MOUNT_ROOT}/${jfmc} and ${JFMC_MOUNT_ROOT}/${installer} failed\nPlease run the following command on host to fix this : mkdir -p ${JFMC_USER_ID}:${JFMC_USER_GROUP_ID} ${JFMC_MOUNT_ROOT}/${jfmc} ${JFMC_MOUNT_ROOT}/${installer} && chown -R ${JFMC_USER_ID}:${JFMC_USER_GROUP_ID} ${JFMC_MOUNT_ROOT}/${jfmc} ${JFMC_MOUNT_ROOT}/${installer}"
    log "Setting ownership on ${JFMC_MOUNT_ROOT}/${installer} and ${JFMC_MOUNT_ROOT}/${jfmc} on host"
    chown -R ${JFMC_USER_NAME}:${JFMC_USER_GROUP} ${installer_mount_root} ${jfmc_mount_root} || exitOnError "Setting ownership on following directories failed : ${JFMC_MOUNT_ROOT}/${jfmc} and ${JFMC_MOUNT_ROOT}/${installer} failed\nPlease run the following command on host to fix this : chown -R ${JFMC_USER_ID}:${JFMC_USER_GROUP_ID} ${JFMC_MOUNT_ROOT}/${jfmc} ${JFMC_MOUNT_ROOT}/${installer}"
}

updateCallHomeInfo() {
    # will generate date of format : 2018-03-08T15:34:04Z
    JFMC_CREATE_DATE=$(date --utc +%FT%TZ)
    addToEnvFile "JFMC_CREATE_DATE"
}

setHostIP() {
    JFMC_HOST_IP=${JFMC_HOST_IP:-127.0.0.1}
    
    if ! isESInstalled; then
        return
    fi

    if ! isHAInstall; then
        return 
    fi

    if [[ $USE_DEFAULTS == "true" ]]; then
        if [[ -z "${JFMC_HOST_IP}" ]] || [[ "${JFMC_HOST_IP}" == "127.0.0.1" ]]; then
            exitOnError "Unable to get ip or dns of this host, please set JFMC_HOST_IP as environment variable and try again"
        fi
    else
        getUserInput "Enter ip or dns of this host, this will be used to connect from other nodes" "${JFMC_HOST_IP}"
        export JFMC_HOST_IP=${users_choice}
    fi
    addToEnvFile "JFMC_HOST_IP"
}

##################################
############## MAIN ##############
##################################

main() {
    aparse $@
    migrateInstallInfo
    isCommandAllowed ${command}
    addToEnvFile "JFMC_HOSTNAME"

    case $command in
        installLocal)
            ( setVersions && \
              log "\nINFO: Mission Control was successfully installed. Run: ./mission-control start"
            ) || exitOnError "Installation failed"
            ;;
        install)
            DATE=$(date +"%Y%m%d%H%M")
            INSTALLATION_LOG_FILE="${INSTALLER_DATA_FOLDER}/installation.${DATE}.log"
            ( 
              setClusterConfig && \
              chooseTypeOfInstallation && \
              setESEnv && \
              runDockerCommand pull && \
              setVersions && \
              removeExposedPorts && \
              setupESHAMount && \
              setupDirsPermissions && \
              copySeedData && \
              cp docker-compose.json ${INSTALLER_DATA_FOLDER}/docker-compose.json && \
              updateCallHomeInfo && \
              log "\nINFO: Mission Control ${MAINTAINER_VERSION} was successfully installed. Run: ./mission-control start" \
            ) | tee -a $INSTALLATION_LOG_FILE || exitOnError "Installation failed. Logs can be found at $DISPLAY_INSTALL_LOG_LOCATION"
            ;;
        upgrade)
            ( 
              backupComposeFile && \
              chooseTypeOfUpgrade && \
              addMongoMigrationDetails && \
              runDockerCommand pull && \
              runDockerCommand stopApp && \
              backwardCmpatibleProcess && \
              setVersions && \
              removeExposedPorts && \
              setupESHAMount && \
              setupDirsPermissions && \
              copySeedData && \
              cp docker-compose.json ${INSTALLER_DATA_FOLDER}/docker-compose.json && \
              updateCallHomeInfo && \
              log "\nINFO: Mission Control was successfully upgraded to ${MAINTAINER_VERSION}. Run: ./mission-control start" \
            ) || exitOnError "Upgrade failed"
            ;;
        upgradeLocal)
            ( runDockerCommand stopApp && \
              setVersions && \
              log "\nINFO: Mission Control was successfully upgraded. Run: ./mission-control start" \
            ) || exitOnError "Upgrade failed"
            ;;
        start)
            runDockerCommand "up -d --remove-orphans" && \
            setupDirsPermissions && \
            configureDatabases && \
            healthCheck && \
            log "\nINFO: Mission Control was successfully started and can be accessed on port $JFMC_PORT"
            ;;
        ps|logs|stop|kill)
            runDockerCommand ${command}
            ;;
        restart)
            runDockerCommand ${command} && \
            setupDirsPermissions && \
            healthCheck && \
            log "\nINFO: Mission Control was successfully started and can be accessed on port $JFMC_PORT"
            ;;
        version)
            if [ -f $INSTALLED_VERSION_FILE ]; then
                echo "Mission Control version: ${INSTALLED_VERSION}"
            else
                echo "Mission Control is not installed"
            fi
            ;;
        x-compose)
                xComposeCmd="docker-compose -f docker-compose.json -p ${COMPOSE_PROJECT_NAME} ${ALL_ARGS}"
                echo "Advanced user: Going to execute '${xComposeCmd}'"
                ${xComposeCmd}
            ;;
        info)
               echo "Mission Control info:"
               echo "project name: ${COMPOSE_PROJECT_NAME}"
               echo "docker repo: ${JFMC_DOCKER_REPO}"
               echo "docker mount folder: ${JFMC_MOUNT_ROOT}"
            ;;
        status)
                healthCheck
            ;;
        *)
            usage
            ;;
    esac
}

main $@
